use game::Game;
use evaluate::PosValue;
use std::collections::{HashMap};
use log::info;
use zobrist;

use crate::{movegen::Move, zobrist::Hash};

#[derive(Clone, PartialEq)]
pub enum EntryType {
    Value,
    LowerBound,
    UpperBound,
}

#[derive(Clone)]
pub struct CacheEntry {
    pub entry_type: EntryType,
    pub halfmove_age: u16,
    pub mov: Move,
    pub depth: i16,
    pub value: PosValue,
}

impl CacheEntry {
    pub fn null() -> Self {
        CacheEntry { entry_type: EntryType::Value, halfmove_age: 0, mov: Move::Nullmove, depth: 0, value: 0 }
    }

    pub fn new_value(depth: i16, halfmove_age: u16, mov: Move, value: PosValue) -> Self {
        CacheEntry {
            entry_type: EntryType::Value,
            halfmove_age,
            mov,
            depth,
            value,
        }
    }

    pub fn new_upper(depth: i16, halfmove_age: u16, mov: Move, value: PosValue) -> Self {
        CacheEntry {
            entry_type: EntryType::UpperBound,
            halfmove_age,
            mov,
            depth,
            value,
        }
    }

    pub fn new_lower(depth: i16, halfmove_age: u16, mov: Move, value: PosValue) -> Self {
        CacheEntry {
            entry_type: EntryType::LowerBound,
            halfmove_age,
            mov,
            depth,
            value,
        }
    }
}

type TableEntry = (zobrist::Hash, CacheEntry);
pub struct Cache {
    table: Vec<TableEntry>,
}

#[derive(Clone)]
pub struct RepetitionTable {
    hashmap: HashMap<zobrist::Hash, i32>,
}

impl Cache {
    pub fn new() -> Self {
        let mut c = Cache {
            table: Vec::new(),
        };
        c.table.resize(4096 * 512, (0, CacheEntry::null()));
        c
    }

    fn get_index(&self, hash: zobrist::Hash) -> usize {
        (hash % (self.table.len() as u64)) as usize
    }

    pub fn lookup<'a>(&'a self, game_pos: &Game) -> Option<CacheEntry> {
        if game_pos.zobrist.is_none() {
            info!("invalid zobrist");
        }
        let hash = game_pos.zobrist.as_ref().unwrap().1;
        let index = self.get_index(hash);
        let table_entry = &self.table[index];

        if table_entry.0 == hash {
            Some(table_entry.1.clone())
        }
        else {
            None
        }
    }

    fn should_replace(old: &TableEntry, new_hash: Hash, new: &CacheEntry) -> bool {
        let (old_hash, old_ce) = old;

        if *old_hash == 0 {
            return true;
        }

        // different positions
        if *old_hash != new_hash {
            if old_ce.halfmove_age < new.halfmove_age {
                return true;
            }
            if old_ce.entry_type == EntryType::Value {
                return false;
            }
            if old_ce.depth <= new.depth {
                return true;
            }
        }

        if new.depth >= old_ce.depth {
            return true;
        }

        return false;
    }
    
    pub fn cache(&mut self, game_pos: &Game, ce: CacheEntry) {
        if game_pos.zobrist.is_none() {
            info!("invalid zobrist");
        }

        let hash = game_pos.zobrist.as_ref().unwrap().1;
        let index = self.get_index(hash);
        let existing_entry = &self.table[index];

        if Self::should_replace(existing_entry, hash, &ce) {
            self.table[index] = (hash, ce);
        }
    }

    pub fn fullness_permill(&self) -> usize{
        let mut permill = 0;
        for i in 0..100 {
            if self.table[i].0 != 0 {
                permill += 10;
            }
        }
        permill
    }
}


impl RepetitionTable {
    pub fn new() -> Self {
        RepetitionTable {
            hashmap: HashMap::with_capacity(1024)
        }
    }

    pub fn clear(&mut self) {
        self.hashmap.clear();
    }

    pub fn increment(&mut self, hash: zobrist::Hash) -> i32 {
        if let Some(entry) = self.hashmap.get_mut(&hash) {
            *entry += 1;
            *entry
        }
        else {
            self.hashmap.insert(hash, 1);
            1
        }
    }

    pub fn decrement(&mut self, hash: zobrist::Hash) {
        if let Some(entry) = self.hashmap.get_mut(&hash) {
            *entry -= 1;
            if *entry == 0 {
                self.hashmap.remove(&hash);
            }
        }
    }
}

